# -*- coding: utf-8 -*-
#
# Copyright (C) 2009-2011 UNINETT AS
#
# This file is part of Network Administration Visualized (NAV).
#
# NAV is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 2 as published by the Free
# Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.  You should have received a copy of the GNU General Public License
# along with NAV. If not, see <http://www.gnu.org/licenses/>.
#
"""Navbar (tools, preferences) and login related controllers"""


import os
from datetime import datetime
import logging
from operator import attrgetter

from django.core.urlresolvers import reverse
from django.http import HttpResponseRedirect, HttpResponse, JsonResponse
from django.views.decorators.http import require_POST
from django.views.decorators.debug import (sensitive_variables,
                                           sensitive_post_parameters)
from django.shortcuts import get_object_or_404, render

from nav.django.auth import ACCOUNT_ID_VAR, desudo
from nav.path import sysconfdir
from nav.django.utils import get_account
from nav.models.profiles import NavbarLink, AccountDashboard, AccountNavlet
from nav.web import ldapauth, auth
from nav.web.utils import require_param
from nav.web.webfront.utils import quick_read, tool_list
from nav.web.webfront.forms import (LoginForm, NavbarLinkFormSet,
                                    ChangePasswordForm, ColumnsForm)
from nav.web.navlets import list_navlets
from nav.web.message import new_message, Messages
from nav.web.webfront import get_widget_columns, find_dashboard

_logger = logging.getLogger('nav.web.tools')

WEBCONF_DIR_PATH = os.path.join(sysconfdir, "webfront")
WELCOME_ANONYMOUS_PATH = os.path.join(WEBCONF_DIR_PATH, "welcome-anonymous.txt")
WELCOME_REGISTERED_PATH = os.path.join(WEBCONF_DIR_PATH,
                                       "welcome-registered.txt")
NAV_LINKS_PATH = os.path.join(WEBCONF_DIR_PATH, "nav-links.conf")


def index(request, did=None):
    """Controller for main page."""
    # Read files that will be displayed on front page
    if request.account.is_default_account():
        welcome = quick_read(WELCOME_ANONYMOUS_PATH)
    else:
        welcome = quick_read(WELCOME_REGISTERED_PATH)

    dashboard = find_dashboard(request.account, did)
    dashboards = AccountDashboard.objects.filter(account=request.account)

    context = {
        'navpath': [('Home', '/')],
        'date_now': datetime.today(),
        'welcome': welcome,
        'dashboard': dashboard,
        'dashboards': dashboards.exclude(id=dashboard.pk),
        'navlets': list_navlets(),
        'title': u'NAV - {}'.format(dashboard.name),
    }

    if dashboards.count() > 1:
        dashboard_ids = [d.pk for d in dashboards]
        current_index = dashboard_ids.index(dashboard.pk)
        previous_index = current_index - 1
        next_index = current_index + 1
        if current_index == len(dashboard_ids) - 1:
            next_index = 0
        context.update({
            'previous_dashboard': dashboards.get(pk=dashboard_ids[previous_index]),
            'next_dashboard': dashboards.get(pk=dashboard_ids[next_index])
        })

    return render(
        request,
        'webfront/index.html',
        context
    )

@sensitive_post_parameters('password')
def login(request):
    """Controller for the login page"""
    if request.method == 'POST':
        return do_login(request)

    origin = request.GET.get('origin', '').strip()
    if 'noaccess' in request.GET:
        if request.account.is_default_account():
            errors = ['You need to log in to access this resource']
        else:
            errors = ['You have insufficient privileges to access this '
                      'resource. Please log in as another user.']
    else:
        errors = []

    return render(
        request,
        'webfront/login.html',
        {
            'form': LoginForm(initial={'origin': origin}),
            'origin': origin,
            'errors': errors,
        }
    )


@sensitive_variables('password')
def do_login(request):
    """Do a login based on post parameters"""
    errors = []
    form = LoginForm(request.POST)
    origin = request.POST.get('origin', '').strip()

    if form.is_valid():
        username = form.cleaned_data['username']
        password = form.cleaned_data['password']

        try:
            account = auth.authenticate(username, password)
        except ldapauth.Error, error:
            errors.append('Error while talking to LDAP:\n%s' % error)
        else:
            if account:
                try:
                    request.session[ACCOUNT_ID_VAR] = account.id
                    request.account = account
                except ldapauth.Error, error:
                    errors.append('Error while talking to LDAP:\n%s' % error)
                else:
                    _logger.info("%s successfully logged in", account.login)
                    if not origin:
                        origin = reverse('webfront-index')
                    return HttpResponseRedirect(origin)
            else:
                _logger.info("failed login: %r", username)
                errors.append('Username or password is incorrect, or the '
                              'account is locked.')

    # Something went wrong. Display login page with errors.
    return render(
        request,
        'webfront/login.html',
        {
            'form': form,
            'errors': errors,
            'origin': origin,
        }
    )


def logout(request):
    """Controller for doing a logout"""
    if request.method == 'POST' and 'submit_desudo' in request.POST:
        desudo(request)
        return HttpResponseRedirect(reverse('webfront-index'))
    else:
        del request.session[ACCOUNT_ID_VAR]
        del request.account
        request.session.set_expiry(datetime.now())
        request.session.save()
    return HttpResponseRedirect('/')


def about(request):
    """Controller for the about page"""
    return render(
        request,
        'webfront/about.html',
        {
            'navpath': [('Home', '/'), ('About', None)],
            'title': 'About NAV',
        }
    )


def toolbox(request):
    """Render the toolbox"""
    account = request.account
    tools = sorted(tool_list(account), key=attrgetter('name'))

    return render(
        request,
        'webfront/toolbox.html',
        {
            'navpath': [('Home', '/'), ('Toolbox', None)],
            'tools': tools,
            'title': 'NAV toolbox',
        },
    )


def _create_preference_context(request):
    """
    Creates a context used by different views for the multiform preference page
    """
    account = get_account(request)

    if account.ext_sync:
        password_form = None
    else:
        password_form = ChangePasswordForm()

    context = {
        'navpath': [('Home', '/'), ('Preferences', None)],
        'title': 'Personal NAV preferences',
        'password_form': password_form,
        'columns_form': ColumnsForm(
            initial={'num_columns': get_widget_columns(account)}),
        'account': account,
        'tool': {'name': 'My account',
                 'description': 'Edit my personal NAV account settings'},
        'navbar_formset': NavbarLinkFormSet(
            queryset=NavbarLink.objects.filter(account=account)),
    }

    return context


def preferences(request):
    """ My preferences """
    context = _create_preference_context(request)

    return render(
        request,
        'webfront/preferences.html',
        context
    )


@sensitive_post_parameters('old_password', 'new_password1', 'new_password2')
def change_password(request):
    """ Handles POST requests to change a users password """
    context = _create_preference_context(request)
    account = get_account(request)

    if account.is_default_account():
        return render(request, 'useradmin/not-logged-in.html', {})

    if request.method == 'POST':
        password_form = ChangePasswordForm(request.POST, my_account=account)

        if password_form.is_valid():
            account.set_password(password_form.cleaned_data['new_password1'])
            account.save()
            new_message(request, 'Your password has been changed.',
                        type=Messages.SUCCESS)
        else:
            context['password_form'] = password_form
            return render(
                request,
                'webfront/preferences.html',
                context
            )

    return HttpResponseRedirect(reverse('webfront-preferences'))


def save_links(request):
    """ Saves navigation preference links on a user """
    account = get_account(request)
    context = _create_preference_context(request)

    if request.method == 'POST':
        formset = NavbarLinkFormSet(request.POST)
        if formset.is_valid():
            for form in formset.deleted_forms:
                instance = form.save(commit=False)
                instance.account = account
                instance.save()

            instances = formset.save(commit=False)
            for instance in instances:
                instance.account = account
                instance.save()
            new_message(request, 'Your links were saved.',
                        type=Messages.SUCCESS)
        else:
            context['navbar_formset'] = formset

            return render(
                request,
                'webfront/preferences.html',
                context
            )

    return HttpResponseRedirect(reverse('webfront-preferences'))


def set_widget_columns(request):
    """Set the number of columns on the webfront"""
    if request.method == 'POST':
        form = ColumnsForm(request.POST)
        if form.is_valid():
            account = request.account
            num_columns = form.cleaned_data.get('num_columns')
            account.preferences[
                account.PREFERENCE_KEY_WIDGET_COLUMNS] = num_columns
            account.save()
            return HttpResponseRedirect(reverse('webfront-index'))
    return HttpResponseRedirect(reverse('webfront-preferences'))


def set_account_preference(request):
    """Set account preference using url attributes"""
    account = request.account
    account.preferences.update(request.GET.dict())
    account.save()
    return HttpResponse()


@require_POST
def set_default_dashboard(request, did):
    """Set the default dashboard for the user"""
    dash = get_object_or_404(AccountDashboard, pk=did)
    try:
        old_default = AccountDashboard.objects.get(account=request.account,
                                                   is_default=True)
    except AccountDashboard.DoesNotExist:
        # No previous default
        old_default = None

    dash.is_default = True
    dash.save()
    if old_default:
        old_default.is_default = False
        old_default.save()
    return HttpResponse(
        u'Default dashboard set to «{}»'.format(dash.name))


@require_POST
def add_dashboard(request):
    """Add a new dashboard to this user"""
    name = request.POST.get('dashboard-name', 'New dashboard')
    dashboard = AccountDashboard(account=request.account, name=name)
    dashboard.save()
    return JsonResponse({'dashboard_id': dashboard.pk})


@require_POST
def delete_dashboard(request, did):
    """Delete this dashboard and all widgets on it"""
    is_last = AccountDashboard.objects.filter(
        account=request.account).count() == 1
    if is_last:
        return HttpResponse('Can not delete last dashboard', status=400)

    dash = get_object_or_404(AccountDashboard, pk=did,
                             account=request.account)
    dash.delete()

    return HttpResponse('Dashboard deleted')


@require_POST
def rename_dashboard(request, did):
    """Rename this dashboard"""
    dash = get_object_or_404(AccountDashboard, pk=did)
    dash.name = request.POST.get('dashboard-name', dash.name)
    dash.save()
    return HttpResponse(
        u'Dashboard renamed to «{}»'.format(dash.name))


@require_POST
def save_dashboard_columns(request, did):
    """Save the number of columns for this dashboard"""

    # Explicit fetch on account to prevent other people to change settings
    dashboard = get_object_or_404(AccountDashboard, pk=did,
                                  account=request.account)
    dashboard.num_columns = request.POST.get('num_columns', 3)
    dashboard.save()
    return HttpResponse()


@require_POST
@require_param('widget_id')
def moveto_dashboard(request, did):
    """Move a widget to this dashboard"""
    account = request.account
    dashboard = get_object_or_404(AccountDashboard, account=account, pk=did)
    widget = get_object_or_404(AccountNavlet, account=account,
                      pk=request.POST.get('widget_id'))
    widget.dashboard = dashboard
    widget.save()
    return HttpResponse(u'Widget moved to {}'.format(dashboard))
